"""lemming.py

Part of the Tundra project.
(C) 2006 Rebuild Software
http://rebui.ld.net.au

This file contains the base lemming class, manipulatable by skill scripts (see "skills/s_*.py")
"""

import globals as g
from excepts import *

import pygame
from pygame.locals import *

from math import *

import cache

class Lemming(pygame.sprite.Sprite):
    """Main class for controlling a single lemming."""

    # NOTE: If the lemming moves in a staggery diagonal line from the entrance hatch
    # without touching anything, this indicates the game.collisionMask surface hasn't
    # been initialized properly
    
    def __init__(self,(x,y),game):
        """Lemming((x,y),game) -> Creates a new lemming object at level position (x,y), reading from surf_coll"""
        pygame.sprite.Sprite.__init__(self)

        self.MAX_FALL = 60
        self.MAX_BOUND_HEIGHT = 6
        
        self.isFloater = False
        self.isClimber = False
        self.isBomber = False
                
        self.isAlmostDead = False
        self.isAlmostDeadTimer = 0

        self.posX = x
        self.posY = y
        
        self.direction = 1 # -1 == left, 1 == right
        self.currentSkill = "s_faller"
        self.currentFrame = 0
        self.altitude = 0
        self.friendlyName = None
        self.game = game

        self.countdown = None

        self.rect = Rect(g.skillCache.frames(self.currentSkill)[self.direction][1])
        self.frameCount = len(g.skillCache.frames(self.currentSkill)[self.direction][0])
        
        return
        
    def update(self):
        """l.update() ->  advances movement by one frame, depending on current action"""
        if self.currentSkill == "s_ghost":
            if self.isAlmostDeadTimer <= 0:
                self.kill()
            else:
                self.isAlmostDeadTimer -= 1        
        else:
            #instakill conditions:
            try:
                #turner conditions
                if self.checkPos((self.posX, self.posY), "turnLeft"):
                    self.direction = -1
                elif self.checkPos((self.posX, self.posY), "turnRight"):
                    self.direction = 1

                if self.isBomber == True:  # the countdown timer has run out
                    self.isBomber = False
                    self.setSkill("s_goner", ignoreDeathAction=True)
                    return
                elif self.checkPos((self.posX, self.posY+1), "water"):
                    if not self.currentSkill == "s_drowner":
                        self.setSkill("s_drowner", ignoreDeathAction=True)
                        return
                elif self.checkPos((self.posX, self.posY+4), "hazard"): 
                # due to the way active rectangles are stored, it's a y offset of 4
                    if not self.currentSkill == "s_burner":
                        self.setSkill("s_burner", ignoreDeathAction=True)
                        return
                elif self.checkPos((self.posX, self.posY+4), "trap"):
                    for x in self.game.trapRects:
                        if x[0].collidepoint((self.posX, self.posY+4)) and not x[1].animating:
                            x[1].trigger()
                            self.killDelay(x[1].frameCount)                            
                            return
            except OutOfThePlayfield:
                pass
        
            g.skillCache.get(self.currentSkill, self)  
        return

    def setSkill(self, name, ignoreDeathAction=False):
        """l.setSkill(name) -> changes lemming skill (if possible)"""
        if g.skillCache.check(self.currentSkill, "isDeathAction") and not ignoreDeathAction: 
            # don't let people allocate skills to lemmings that are about to meet their maker
            raise SkillUnsettable
        elif g.skillCache.check(name, "mustStartOnGround") and not g.skillCache.check(self.currentSkill, "isOnGround"):
            raise SkillUnsettable
        
        if not g.skillCache.check(name, "propertyChangeOnly"):  # skill not actually going to change, don't bother cleaning up
            g.skillCache.cleanUp(self.currentSkill, self)

        g.skillCache.get(name, self)
        return

    def checkPos(self,(x,y),attr):
        try:
            result = self.game.collisionMask.get_at((x, y))
            value = g.MAP_KEY.get(attr,(0,0))
            return result[value[0]]&value[1] == value[1]
        except IndexError:
            raise OutOfThePlayfield # run away! run away!

    def getDisplayName(self):
        if not self.friendlyName:
            return self.currentSkill
        else:
            return self.friendlyName

    def _incAnimFrame(self):
        """l._incAnimFrame() -> cycle animation of current skill in the appropriate manner"""
        if self.currentFrame >= self.frameCount:
            self.currentFrame = 0
        self.image = g.skillCache.frames(self.currentSkill)[self.direction][0][self.currentFrame]
        self.rect.midbottom = (self.posX, self.posY+1)
        self.currentFrame += 1
        return

    def killDelay(self, delay):
        self.isAlmostDeadTimer = delay
        self.setSkill("s_ghost", ignoreDeathAction=True)

    def kill(self):
        """l.kill() -> deconstructs this lemming"""
        self.game.protagonists.remove(self)
        if self.countdown:
            self.countdown.kill()
        self.game.numLemmingsReleased -= 1
        return
    
    def explode(self):
        self.isBomber = True    # intentionally no setSkill code is used here. 
                                # this flag tells Lemming.update() to setSkill("s_goner"), so that 
                                # game.upperAnims (which controls the countdown, which calls this method ) 
                                # doesn't skip a frame of animation.
        return

    def save(self):
        """l.save() -> decostructs this lemming, whilst improving one's score"""
        self.game.protagonists.remove(self)
        self.game.numLemmingsReleased -= 1
        self.game.numLemmingsSaved += 1
        return


# note: func is a lambda function that takes lemming as an arg
class Countdown(pygame.sprite.Sprite):
    def __init__(self, lemming, ms, func):
        pygame.sprite.Sprite.__init__(self)
        self.lemming = lemming
        self.func = func
        self.numFrames = abs(ms)*g.FPS/1000
        self.index = (self.numFrames / g.FPS)
        self.image = g.countdownCache.get(self.index)
        self.rect = self.image.get_rect()
        self.rect.midbottom = (self.lemming.posX+1+self.lemming.direction, self.lemming.posY-12)

        self.lemming.game.upperAnims.add(self)   # this handles both the update() and draw() functions

    def update(self):
        if self.lemming.countdown == None:
            self.lemming.game.upperAnims.remove(self)
            self.lemming = None
            return
            
        if (self.numFrames % g.FPS) == 0:
            if self.index <= 0:
                try:                
                    self.func(self.lemming)
                except SkillUnsettable:
                    pass
                self.kill()
                return
            self.index -= 1
            self.image = g.countdownCache.get(self.index+1)
            self.rect = self.image.get_rect()

        self.rect.midbottom = (self.lemming.posX+1+self.lemming.direction, self.lemming.posY -12 )
        self.numFrames -= 1
        return

    def kill(self):
        self.lemming.countdown = None
        self.image = pygame.Surface([6,6]).convert_alpha()
        self.image.fill((0,0,0,0))
        return
